#include <nds.h>
#include <fat.h>

#include <stdio.h>

#include "font.h"

#define SC_MODE_REG 			*(vu16*)0x09FFFFFE
#define SC_MODE_MAGIC			(u16)0xA55A
#define SC_MODE_FLASH_RW		(u16)0x4
#define SC_MODE_FLASH_RW_LITE	(u16)0x1510


#define SC_FLASH_MAGIC_ADDR_1	(*(vu16*) 0x08000b92)
#define SC_FLASH_MAGIC_ADDR_2	(*(vu16*) 0x0800046c)
#define SC_FLASH_MAGIC_1		((u16) 0xaa)
#define SC_FLASH_MAGIC_2		((u16) 0x55)
#define SC_FLASH_ERASE			((u16) 0x80)
#define SC_FLASH_ERASE_BLOCK	((u16) 0x30)
#define SC_FLASH_ERASE_CHIP		((u16) 0x10)
#define SC_FLASH_PROGRAM		((u16) 0xA0)
#define SC_FLASH_IDLE			((u16) 0xF0)
#define SC_FLASH_IDENTIFY		((u16) 0x90)

extern u16 scfw_bin[];
extern u16 scfw_binEnd[];

static PrintConsole tpConsole;
static PrintConsole btConsole;

static int bg;
static int bgSub;

u32 sc_flash_id() {
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_MAGIC_1;
	SC_FLASH_MAGIC_ADDR_2 = SC_FLASH_MAGIC_2;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_IDENTIFY;
	
	// should equal 0x000422b9
	u32 res = SC_FLASH_MAGIC_ADDR_1;
	res |= *GBA_BUS << 16;
	
	*GBA_BUS = SC_FLASH_IDLE;
	
	return res;

}

void sc_flash_rw_enable() {
	bool buf = REG_IME;
	REG_IME = 0;
	SC_MODE_REG = SC_MODE_MAGIC;
	SC_MODE_REG = SC_MODE_MAGIC;
	SC_MODE_REG = SC_MODE_FLASH_RW;
	SC_MODE_REG = SC_MODE_FLASH_RW;
	REG_IME = buf;
}

void sc_flash_erase_chip() {
	bool buf = REG_IME;
	REG_IME = 0;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_MAGIC_1;
	SC_FLASH_MAGIC_ADDR_2 = SC_FLASH_MAGIC_2;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_ERASE;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_MAGIC_1;
	SC_FLASH_MAGIC_ADDR_2 = SC_FLASH_MAGIC_2;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_ERASE_CHIP;
	
	while (*GBA_BUS != *GBA_BUS)swiWaitForVBlank();
	*GBA_BUS = SC_FLASH_IDLE;
	REG_IME = buf;
}

void sc_flash_erase_block(vu16 *addr) {
	bool buf = REG_IME;
	REG_IME = 0;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_MAGIC_1;
	SC_FLASH_MAGIC_ADDR_2 = SC_FLASH_MAGIC_2;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_ERASE;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_MAGIC_1;
	SC_FLASH_MAGIC_ADDR_2 = SC_FLASH_MAGIC_2;
	*addr = SC_FLASH_ERASE_BLOCK;
	
	while (*GBA_BUS != *GBA_BUS)swiWaitForVBlank();
	*GBA_BUS = SC_FLASH_IDLE;
	REG_IME = buf;
}

void sc_flash_program(vu16 *addr, u16 val) {
	bool buf = REG_IME;
	REG_IME = 0;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_MAGIC_1;
	SC_FLASH_MAGIC_ADDR_2 = SC_FLASH_MAGIC_2;
	SC_FLASH_MAGIC_ADDR_1 = SC_FLASH_PROGRAM;
	*addr = val;
	
	while (*GBA_BUS != *GBA_BUS)swiWaitForVBlank();
	
	*GBA_BUS = SC_FLASH_IDLE;
	REG_IME = buf;
}


void CustomConsoleInit() {
	videoSetMode(MODE_0_2D);
	videoSetModeSub(MODE_0_2D);
	vramSetBankA (VRAM_A_MAIN_BG);
	vramSetBankC (VRAM_C_SUB_BG);
	
	bg = bgInit(3, BgType_Bmp8, BgSize_B8_256x256, 1, 0);
	bgSub = bgInitSub(3, BgType_Bmp8, BgSize_B8_256x256, 1, 0);
		
	consoleInit(&btConsole, 3, BgType_Text4bpp, BgSize_T_256x256, 20, 0, false, false);
	consoleInit(&tpConsole, 3, BgType_Text4bpp, BgSize_T_256x256, 20, 0, true, false);
		
	ConsoleFont font;
	font.gfx = (u16*)fontTiles;
	font.pal = (u16*)fontPal;
	font.numChars = 95;
	font.numColors =  fontPalLen / 2;
	font.bpp = 4;
	font.asciiOffset = 32;
	font.convertSingleColor = true;
	consoleSetFont(&btConsole, &font);
	consoleSetFont(&tpConsole, &font);

	consoleSelect(&tpConsole);
}

bool Prompt() {
	while(1) {
		swiWaitForVBlank();
		scanKeys();
		switch (keysDown()) {
			case KEY_A: return true;
			case KEY_B: return false;
		}
	}
}


int main(void) {
	CustomConsoleInit();
	
	sysSetCartOwner(true);
		
	sc_flash_rw_enable();
	printf("\n\n\n\n\n\n\n\n\n\n\n");
	iprintf("        Flash ID %lx", sc_flash_id());
	consoleSelect(&btConsole);
	printf("\n Press [A] to kill supercard.\n");
	printf(" Press [B] to spare supercard.\n");
	
	if (!Prompt())return 0;
	
	consoleClear();
	
	printf("      Death 2 supercard :3\n");
	
	printf("      Erasing whole chip\n");
	sc_flash_erase_chip();
	printf("      Erased whole chip\n");
	
	for (u32 off = 0; off < (u32)(scfw_binEnd - scfw_bin); off += 2) {
		u16 val = 0;
		val |= scfw_bin[off];
		val |= scfw_bin[off+1] << 8;
		sc_flash_program((vu16*)(0x08000000 + off), val);
		if (!(off & 0x00ff))iprintf("      Programmed %lx\n", 0x08000000 + off);
	}
	
	printf("      Ded!\n");
	 
	while(1) {
		swiWaitForVBlank();
		scanKeys();
		if (keysDown())return 0;
	}
	return 0;
}

